(*****************************************************************
This assignment changes the grammar from the last one to add types.
To simplify the assignment, the grammar has also been changed so
that user-defined functions take only one argument.

The grammar for the language is given below:


Z ::=  ... -1 | 0 | 1 | ...
B ::=  true | false
p ::=  > | < | = | && | || | * | + | -
x,y,.. \in Var

ty :: int | bool | ty -> ty | (ty * ty)

e ::=  Z | B | x |
       | p(e1,...en)
       | fun (x:ty):ty => e
       | let x:ty = e in e
       | rec f(x:ty):ty => e in e
       | if e then e else e
       | (e e)
       | (e)

 **********************************************************************)


module Exp = struct
  type var = string 
  type funName = string
  type arg = string

  type name = Var of var | Fun of funName

  type valueKind = Fun | Var | Rec

  type prim = Plus | Minus | Times | Equal | Lt | Gt | And | Or

  (* There are two base types in the language - int and bool.  Additionally,
     users can write functions that are typed using the "->" (arrow) constructor.
     The type of primitive operations (+, -, ...) use a tuple type to represent
     their arguments.  A type variable represents an "unknown" type - it will
     be used in the implementation of the type inference procedure. *)

  type ty = Int | Bool | Arrow of (ty * ty) | Tuple of (ty list) | TyVar of string

  type tyarg = (arg * ty)

  (* The grammar allows you to write both typed and untyped expressions.  You use
     the procedure typeCheck to check the former, and the procedure inferType to
     infer types for the latter. *)
               
  type value = I of int | B of bool |  TyFn of (tyarg * ty * t) | TyRec of (funName * tyarg * ty * t) | Fn of (arg * t) | Rec of (funName * arg * t)

  and t =
      Const of value
    | Var of var
    | PrimCall of (prim * t list)
    | TyBind of (tyarg * t * t)
    | TyBindRec of (funName * tyarg * ty * t * t)
    | Bind of (var * t * t)
    | BindRec of (funName * arg * t * t)
    | Call of (t * t)
    | If of (t * t * t) 
end

module rec TypeCheck :
sig

  val typeCheck : Exp.t -> Exp.ty
  val inferType : Exp.t -> Exp.ty
  
end =
struct
  open Exp
  exception BindingNotFound of string
  exception PrimNotFound of string
  exception IllTypedExpression of string
  exception Unhandled of string
  exception FailedInference of string

  (* A map that relates variables to their types *)
  type gamma = Exp.var -> Exp.ty

  (* A map that relates type variables to their types 
     if such a type has been inferred *)
                            
  type constraintMap = Exp.ty -> Exp.ty option

  (* The inferred type of an expression is a type that
     potentially includes type variables whose value
     is provided by a constraintMap *)
     
  type inferType = { ty : Exp.ty; c : constraintMap }

  let emptyMap = fun () -> fun s -> raise (BindingNotFound(s))
  let emptyConstraintMap = fun () -> fun s -> None
  
  let lookup m x = m x
      
  let update m x ty =
    fun y -> if (x = y) then ty else (m y)


  let primToString = function
    | Plus -> "plus"
    | Minus -> "minus"
    | Times -> "mult"
    | Equal -> "equals"
    | Lt -> "lessThan"
    | Gt -> "greaterThan"
    | And -> "and"
    | Or ->  "or"

  
  let addPrimTypes (m : gamma) =
    List.fold_left
      (fun m' p ->
         let s = primToString p in
         match s with
         | "plus" ->  update m' s (Arrow ((Tuple [Int;Int]),Int))
         | "minus" -> update m' s (Arrow ((Tuple [Int;Int]),Int))
         | "mult" -> update m' s (Arrow ((Tuple [Int;Int]),Int))
         | "equals" -> update m' s (Arrow ((Tuple [Int;Int]),Bool))
         | "lessThan" -> update m' s (Arrow ((Tuple [Int;Int]),Bool))                        
         | "greaterThan" -> update m' s (Arrow ((Tuple [Int;Int]),Bool))                        
         | "and" -> update m' s (Arrow ((Tuple [Bool;Bool]),Bool))                        
         | "or" -> update m' s (Arrow ((Tuple [Bool;Bool]),Bool))
         | _ -> raise (PrimNotFound(s)))
      m
      [Plus;Minus;Times;Equal;Lt;Gt;And;Or]

  (* Typecheck an expression e given type environment m *)
  let rec check (m : gamma) (e : Exp.t) =
    let m = addPrimTypes m
    in match e with
    (* FILL-IN HERE *)
                  
  let typeCheck e = check (emptyMap())  e


  (* The following functions are used for type inference *)
      
  let genTyVar =
    let c = ref 0 in
    fun () ->
      let r = Exp.TyVar ("X?" ^ (string_of_int !c)) in
      (c := !c + 1; r)

  (* Given an untyped term, return a typed term consisting of
     consistently-defined type variables *)
  let rec annotate (env : string -> Exp.ty) (exp : Exp.t) =
    (* FILL-IN HERE *)

  (* Lookup the type variables that occur in a type *)
  let rec lookupType (c : constraintMap) ty =
    (* FILL-IN HERE *)      
    
  (* Given two types, possibly containing type variables,
     return a new constraintMap that reflects their unification *)
  let rec unify (c : constraintMap) ty1 ty2 : constraintMap =
    (* FILL-IN HERE *)      

  (* Given a typing environment, and constraintMap, return the
     inferred type of an expression *)
  let rec infer (m : gamma) (c : constraintMap) (e : Exp.t) : inferType =
    (* FILL-IN HERE *)      

  let rec simplify ty c =
    match ty with
    | TyVar x ->
      (match (lookupType c ty) with
       | Some ty -> ty
       | None -> ty)
    | Arrow(tyArg,tyRet) ->
      Arrow ((simplify tyArg c),(simplify tyRet c))
    | _ -> ty
    
  let inferType e =
    let gamma = addPrimTypes (emptyMap()) in
    let e' = annotate (gamma) e in
    let {ty = ty; c = c} = infer gamma (emptyConstraintMap()) e' in
    simplify ty c
end
and Print :
sig 
  val printExp : Exp.t -> unit
  val tyrep : Exp.ty -> unit
end =
struct

  open Exp
  open Format

  exception Unhandled of string
      
  let const_int = print_int
  let const_bool = print_bool    
  let kwd = print_string
  let ident = print_string

  let rec tyrep = function
    | Int -> kwd " int "
    | Bool -> kwd " bool "
    | TyVar x -> ident x
    | Arrow (ty1,ty2) -> tyrep ty1; kwd " -> "; tyrep ty2
    | Tuple tyargs ->
      kwd "(";
      let rec f =
        fun l ->
         (match l with
          | [] -> ()
          | hd :: tl -> (tyrep hd; kwd " * "; f tl))
      in f tyargs;
      kwd ")"
        
  let prim = function
    | Plus -> kwd " + "
    | Minus -> kwd " - "
    | Times -> kwd " * "
    | Equal -> kwd " = "
    | Lt -> kwd " < "
    | Gt -> kwd " > "
    | And -> kwd " && "
    | Or -> kwd " ||  "

  let printParams args =
    (kwd "(";
     let rec loop l =
       match l with
       | [] -> kwd ") "
       | [(x,ty)] -> ident x; kwd ":"; tyrep ty
       | x :: l' -> loop [x]; kwd ","; loop l'; kwd ")"
     in loop args)

  let rec printArgs args =
    (kwd "(";
     let rec loop l =
       match l with
       | [] -> kwd ")"
       | [x] -> printExp x; kwd ") "
       | x :: l' -> printExp x; kwd ","; loop l'
     in loop args)
  and printExp = function
    | Const c ->
      (match c with
        | I n -> const_int n
        | B b -> const_bool b
        | TyFn (arg,retType,t) -> (kwd "fn"; printParams [arg]; kwd ":"; tyrep retType; kwd "= "; printExp t)
        | TyRec (f,arg,retType,t) -> (kwd "rec "; ident f; printParams [arg]; kwd ":"; tyrep retType; kwd "= "; printExp t)
        | _ -> raise (Unhandled "unhandled expression in Print"))
    | Var x -> ident x                    
    | PrimCall(p, args) ->
       kwd "("; printExp (List.nth args 0); prim p; printExp (List.nth args 1); kwd ")"
    | TyBind((x,ty),e1,e2) ->
      open_vbox 1; kwd " let "; ident x; kwd " : "; tyrep ty; kwd "= "; printExp e1; print_break 0 0; kwd "in "; printExp e2; close_box()
    | TyBindRec(f,arg,retType,body,e) ->
      open_vbox 1; kwd " rec "; ident f;  printParams [arg]; kwd ":";  kwd ")"; tyrep retType; kwd "= "; printExp body; print_break 0 0; kwd "in "; printExp e; close_box()        
    | Call (t, arg) ->  open_vbox 2; printExp t;  printArgs [arg]; close_box ()
    | If(e1,e2,e3) ->
      open_vbox 2;  kwd "if "; printExp e1; print_break 1 2; 
      kwd " then "; printExp e2; print_break 1 2;
      kwd " else "; printExp e3; close_box ()
    | _ -> raise (Unhandled "Unhandled expression in Print")
end      

  
(**************************** 

Some test cases:

let x:int = 1 in x + 1;;

let f:int->int = fun (x:int):int => x in f 2

let f:int->int = fun (x:int):int => x + 1 in f

let f:(int->int)->int = fun(g:int->int):int => g 10 in f (fun(x:int):int => x)

let f:(int->int)->int = fun(g:int->int):int => g 10 in f (fun(x:int):int => x + 1)

let f:int -> (int -> int) = fun (x:int):int->int => fun(y:int):int => x + y in ((f 2) 3)

rec fact(n:int):int =>  if (n = 0) then 1 else (n * (fact (n - 1))) in (fact 3)

rec fact(n:int):int =>  if (n = 0) then 1 else (n * (fact (n - 1))) in (fact 3)

fun (x:int):int -> int => fun(y:bool):bool => y

let negb:bool->bool = fun(x:bool):bool => if x then false else true in let nandb:bool->bool->bool = fun (a:bool):bool->bool => fun (b:bool):bool => if a then (negb b) else true in (nandb true) true

rec isEven(n:int):bool => if n = 0 then true else if n = 1 then false else isEven (n - 2) in isEven 5

rec sum(n:int):int->int => fun(m:int):int => if n = 0 then m else 1 + ((sum (n - 1)) m) in (sum 2) 3

rec fib(n:int):int => if ((n = 0) || (n = 1)) then 0 else (fib (n - 1)) + (fib (n - 2)) in (fib 5)

let foo = parse("let x:int = 10 in fun (g: int -> bool):int => if (g x) then 1 else 0");;


 ****************************)
